from .main import Subtitle_Edit, sub_args, effects_args, highlight_args, generate_subs_simple
from . import main
from . import Effects
from . import Highlight
from . import utils
from . import docker_wrapper
from .docker_wrapper.base import args_border
from . import handler


import os
from .container import ContainerWrapper
from . import base
import traceback
from pysubs2 import SSAFile
import json


class DockerWrapper(base.BaseWrapper):

    def __init__(self, force_install:bool = False):
        # Check for the install
        super().__init__(force_install)
        self.docker_installed = False

        # Only continue if package is installed
        if self.installed:
            self.image_name = 'n01d3a/aegisub-cli:ShaperyRoundedBorders-2.3'
            self.rep_name = 'n01d3a/aegisub-cli'
            self.client = self.get_client()
            self.image = self.check_image_v2()

    def __call__(self,
                 input_ass:str | SSAFile, # path or SAAFile, not ass in string form
                 args_border:base.args_border, # settings for the args border
                 return_pysubsSSAFile:bool = True, # container_run_func needs to take client and name as a input
                 _traceback:bool = False,
                 verbose:bool = False,
                 cleanup:bool = False
                 ):

        if not self.installed and not self.docker_installed:
            raise ImportError('The Docker Package and Image needs to be installed for this section.')
        # Init variables
        container = None
        container_wrapper = None

        try:
            if args_border.container_run_func is not None:
                container = args_border.container_run_func(self.client, self.image_name)
            else:
                container = self.client.containers.run(
                    self.image_name,
                    command='tail -f /dev/null',
                    detach=True,
                    network_mode='host' #network_mode='host' to prevent an docker error
                )

            container_wrapper = ContainerWrapper(container, verbose)

            # Pre exec

            # Install fonts if they exist and copy input ass
            container_wrapper.copy_needed_files(input_ass, args_border.fonts_path)

            # Install packages (optional)
            if args_border.packages is not None:
                container_wrapper.install_packages(args_border.packages)

            # Execute the shapery command

            dialog_json = json.dumps(
                {"button": 0,
                 "values": args_border()
                 }
            )

            # aegisub-cli --dialog '{"button": 0, "values": {"offset": 20, "radius": 20, "transformY": -1, "heightscaling": 1.3, "borderColor": "&H000000&", "borderAlpha": "&H00"}}' --automation jz.RoundedBorders.lua input.ass output.ass "Create Rounded Border"
            shapery_command_test = 'aegisub-cli --automation ILL.Shapery.moon --loglevel 4 input.ass output.ass ": Shapery macros :/Shape expand" || true'
            shapery_command = ['aegisub-cli', f"--dialog '{dialog_json}'",
                               '--automation', 'jz.RoundedBorders.lua',
                               'input.ass', 'output.ass',
                               '"Create Rounded Border"', '|| true'
                               ]
            container_wrapper(' '.join(shapery_command))

            # Get output from the aegisub-cli
            output : str = container_wrapper.retrieve_file('/home/output.ass', True)
            output : str = output[output.find('[Script Info]'):]

            # make output to sting if wanted and then return it
            if return_pysubsSSAFile:
                output = SSAFile.from_string(output)
            return output
        except Exception as e:
            if _traceback:
                traceback.print_exc()

            # return the output of the even if the program fails

            # convert to pysub2.SAAFile
            if type(input_ass) == str:
                output = SSAFile.load(input_ass)
            else:
                output = input_ass

            # Convert to string if necessary
            if not return_pysubsSSAFile:
                output = output.to_string()
            return output
        finally:
            # Stop and remove the container
            if container is not None:
                if container.status == 'running':
                    container.stop()
                container.remove()

            # stop and remove all other containers from the image if needed
            if cleanup:
                self.cleanup()


    def return_docker_image(self):
        images = self.client.images.list()
        found_image = False
        for image in [x for x in images]:
            tag = image.tags
            if 'n01d3a/aegisub-cli:Shapery' in tag:
                found_image = image
        return found_image

    def ask_install_docker_image(self):

        # Ask for confirmation to download the package
        print('For this Part of the script a specific docker image must be installed')
        answer = ''
        while answer not in ['y', 'n']:
            answer = input('y/n:').lower()

        # Review answer
        if answer == 'y':
            return True
        else:
            return False

    def check_image_v2(self):
        try:
            image = self.client.images.get(self.image_name)
            self.docker_installed = True
        except self.docker.errors.ImageNotFound:

            # ask to install the docker image
            if self.force_install:
                answer = True
            else:
                answer = self.ask_install_docker_image()

            # if yes install docker image and set variable
            if answer:
                self.pull_docker_image()
                image = self.client.images.get(self.image_name)
                self.docker_installed = True
            else:
                # if not, return None Image
                image = None
        return image

    def pull_docker_image(self):
        name, tag = self.image_name.split(':')
        print('Pulling Docker Image...')
        self.client.images.pull(name, tag=tag)
        print('Pulled Docker Image.')

    def cleanup(self):
        containers = self.client.containers.list(all=True)

        # check if image of container can be opened, else delete container
        new_containers = list()
        for container in containers:
            try:
                image = container.image
                new_containers.append(container)
            except self.docker.errors.ImageNotFound:
                container.remove(force=True)
        containers = new_containers

        # stop and remove containers
        for container in containers:
            if container.image == self.image:
                if container.status == 'running':
                    container.stop()
                # try stopping the container just in case
                try:
                    container.stop()
                except Exception:
                    pass
                container.remove()

    def cleanup_old_images(self):
        images = self.client.images.list(self.rep_name)

        # delete all images with different tags
        for image in images:
            tag = image.attrs['RepoTags'][0]
            if not tag == self.image_name:
                self.client.images.remove(tag, force=True)


from pathlib import Path
import pytest
import SubTextHighlight
import pysubs2
import os

UPDATE_GOLDEN = os.environ.get("UPDATE_GOLDEN", "false").lower() == "true"
# Define test cases
TEST_CASES = [
    (
        "one_word_only_and_fade",
        {
            "subtitle_type": 'one_word_only',
            "fill_sub_times": True,
            "word_max": 0,
            "highlight_args": None,
            "effect_args": SubTextHighlight.effects_args(fade=(50, 50)),
        }
    ),
    (
        "separate_on_period_and_highlighting",
        {
            "subtitle_type": 'separate_on_period',
            "fill_sub_times": False,
            "word_max": 11,
            "highlight_args": SubTextHighlight.highlight_args(
                highlight_word_max=0,
                primarycolor='00AAFF'
            ),
            "effect_args": None,
        }
    ),
    (
        "join_and_word_max",
        {
            "subtitle_type": 'join',
            "fill_sub_times": False,
            "word_max": 20,
            "highlight_args": None,
            "effect_args": None,
        }
    ),
    (
        "appear",
        {
            "subtitle_type": 'join',
            "fill_sub_times": False,
            "word_max": 20,
            "highlight_args": None,
            "effect_args": SubTextHighlight.effects_args(fade=(50, 50), appear=True),
        }
    ),
    (
        "rounded_borders",
        {
            "subtitle_type": 'join',
            "fill_sub_times": False,
            "word_max": 20,
            "highlight_args": None,
            "effect_args": SubTextHighlight.effects_args(
                fade=(50, 50),
                args_border=SubTextHighlight.args_border()
            ),
        }
    ),
    (
        "rounded_background_highlight",
        {
            "subtitle_type": 'separate_on_period',
            "fill_sub_times": False,
            "word_max": 11,
            "highlight_args": SubTextHighlight.highlight_args(
                highlight_word_max=0,
            ),
            "effect_args": SubTextHighlight.effects_args(
                fade=(50, 50),
                args_border=SubTextHighlight.args_border(use_borders_as_highlight=True, height_scaling=1.0)
            )
        }
    ),
    (
        "rounded_background_appear",
        {
            "subtitle_type": 'separate_on_period',
            "fill_sub_times": False,
            "word_max": 11,
            "highlight_args": None,
            "effect_args": SubTextHighlight.effects_args(
                fade=(50, 50),
                args_border=SubTextHighlight.args_border(),
                appear=True
            )
        }
    ),
]


@pytest.mark.parametrize("name, options", TEST_CASES)
def test_subtitle(tmp_path, name, options):
    # static paths for testing
    base_path = Path("")
    blank_srt_path = base_path / "input" / "blank.srt"
    video_path = base_path / "input" / "plain_video.mp4"

    # dynamic paths based on the parameters
    expected_ass = base_path / "expected" / (name + '.ass')
    output_ass = base_path / "output" / (name + '.ass')

    sub_args = SubTextHighlight.sub_args(
        input=str(blank_srt_path),
        output=None,
        subtitle_type=options["subtitle_type"],
        fill_sub_times=options["fill_sub_times"],
        alignment=2,
        input_video=str(video_path),
    )

    if options['word_max'] is not None:
        sub_args.word_max = options['word_max']

    sub_edit = SubTextHighlight.Subtitle_Edit(sub_args, options['highlight_args'], options['effect_args'])
    sub_file: pysubs2.SSAFile = sub_edit()

    sub_file.save(str(output_ass))

    # Assert
    if UPDATE_GOLDEN:
        sub_file.save(str(expected_ass))
        assert True
    else:
        # Verify
        actual = sub_file.to_string('ass')
        expected = pysubs2.load(str(expected_ass)).to_string('ass')

        assert actual == expected



import dataclasses
import re
from copy import deepcopy
import pysubs2
import subprocess
import tempfile
import os
import json


def get_duration_resolution(file_path):
    cmd = [
        'ffprobe', '-v', 'error', '-print_format', 'json',
        '-show_format', '-show_streams', file_path
    ]

    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        data = json.loads(result.stdout)
    except (subprocess.CalledProcessError, json.JSONDecodeError):
        return 0.0, None

    # 1. Handle Duration
    # Try getting duration from format first, fallback to 0.0
    duration_str = data.get('format', {}).get('duration', 0)
    duration = float(duration_str)

    # 2. Handle Resolution (Video Only)
    resolution = None
    streams = data.get('streams', [])
    for stream in streams:
        if stream.get('codec_type') == 'video':
            width = stream.get('width')
            height = stream.get('height')
            if width and height:
                resolution = (width, height)
            break

    return duration, resolution

def is_subfile_resolution_set(sub_file:pysubs2.SSAFile):
    info = sub_file.info
    keys = info.keys()
    if 'PlayResX' in keys and 'PlayResY' in keys:
        return True
    else:
        return False

def dprint(txt):
    if os.environ['debug'] == 'True':
        print(txt)

def exec_command(command:list):
    try:
        result = subprocess.run(command, text=True, stdout=subprocess.PIPE)
        if result.returncode != 0:
            raise RuntimeError(result.stdout)
    except Exception as e:
        pass

def add_subtitles_with_ffmpeg(video_path, output_path, sub_file:pysubs2.SSAFile):
    with tempfile.NamedTemporaryFile(mode='w', suffix='.ass', delete=False) as temp_file:
        temp_file.write(sub_file.to_string(format_='ass'))
        temp_filename = temp_file.name
    add_subtitles_with_ffmpeg_with_given_ass(video_path, output_path, temp_filename)
    os.unlink(temp_filename)

def add_subtitles_with_ffmpeg_with_given_ass(video_path, output_path, ass_file):
    command = [
        "ffmpeg",
        "-y",
        "-i", video_path,
        "-vf",
        f"ass={ass_file}",
        "-c:a", "copy",
        "-loglevel", "error",
        output_path
    ]
    exec_command(command)


def hex_to_pysub2_color(hex_color, alpha=0):
    """
    Convert hex color string to pysub2.Color format.

    Args:
        hex_color: A hex string in format 'RRGGBB' (e.g., 'ff0000' for red)
        alpha: Alpha/transparency value (0-255), default 0 (opaque)

    Returns:
        pysub2.Color object
    """

    # Remove '#' if present
    if hex_color.startswith('#'):
        hex_color = hex_color[1:]

    # Convert hex to RGB
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)

    # Create pysub2.Color object
    return pysubs2.Color(r, g, b, alpha)

def import_color(color:pysubs2.Color | str | None):
    if color is None:
        return None
    elif type(color) == pysubs2.Color:
        return color
    else:
        return hex_to_pysub2_color(color)


#@dataclasses.dataclass
@dataclasses.dataclass(kw_only=True)
class args_styles:
    """
            Subtitle style configuration class for customizing text appearance and formatting.

            This class provides comprehensive control over subtitle rendering including font properties,
            colors, visual effects, and layout positioning using SubStation Alpha (SSA/ASS) format standards.

            Parameters:
                fontname (str): Font family name. Any system-installed font can be specified.
                    Default: 'Arial'

                fontsize (float | int): Font size in points. Larger values create bigger text.
                    Default: 24

                primarycolor (pysubs2.Color | str): Main text fill color in RGBA format (0-255).
                    Default: pysubs2.Color(255, 255, 255) (white)

                backcolor (pysubs2.Color | str): Background color behind text when using box border style.
                    Default: pysubs2.Color(0, 0, 0) (black)

                secondarycolor (pysubs2.Color | str): Secondary color for karaoke effects and transitions.
                    Default: pysubs2.Color(0, 0, 0) (black)

                outlinecolor (pysubs2.Color | str): Color of text outline/border for readability.
                    Default: pysubs2.Color(0, 0, 0) (black)

                tertiarycolor (pysubs2.Color | str): Additional outline color for complex border effects.
                    Default: pysubs2.Color(0, 0, 0) (black)

                outline (float | int): Thickness of text outline in pixels. Higher values create thicker borders.
                    Default: 1

                spacing (float | int): Line spacing multiplier. Values <1.0 create tighter spacing, >1.0 looser.
                    Default: 0.75

                shadow (float | int): Drop shadow offset in pixels. 0 disables shadow effect.
                    Default: 0

                alignment (int): Text positioning using numpad layout:
                    1-3: Bottom (left/center/right), 4-6: Middle (left/center/right), 7-9: Top (left/center/right)
                    Default: 5 (middle-center)

                bold (bool): Enable bold text formatting for improved readability.
                    Default: True

                angle (float): Text rotation angle in degrees. Positive values rotate clockwise.
                    Default: 0.0

                borderstyle (int): Border rendering style. 1=outline border, 3=opaque box background.
                    Default: 1

                italic (bool): Enable italic text formatting.
                    Default: False

                underline (bool): Enable underline text formatting.
                    Default: False

            Example:
                >>> # Create style with yellow text and blue outline
                >>> style = args_styles(
                ...     fontsize=28,
                ...     primarycolor=pysubs2.Color(255, 255, 0),
                ...     outlinecolor=pysubs2.Color(0, 100, 255),
                ...     outline=2,
                ...     alignment=2
                ... )

            Note:
                All color parameters accept either pysubs2.Color objects or compatible color strings.
                The default configuration creates bold white text with black outline, optimized for
                readability across various video backgrounds.
            """


    fontname: str = 'Arial'
    fontsize: float | int = 24
    primarycolor: pysubs2.Color | str = dataclasses.field(
        default_factory=lambda: pysubs2.Color(255, 255, 255)
    )
    backcolor: pysubs2.Color | str = dataclasses.field(
        default_factory=lambda: pysubs2.Color(0, 0, 0)
    )
    secondarycolor: pysubs2.Color | str = dataclasses.field(
        default_factory=lambda: pysubs2.Color(0, 0, 0)
    ) # Black for border/shadow
    outlinecolor: pysubs2.Color | str = dataclasses.field(
        default_factory=lambda: pysubs2.Color(0, 0, 0)
    )
    tertiarycolor: pysubs2.Color | str = dataclasses.field(
        default_factory=lambda: pysubs2.Color(0, 0, 0)
    )
    outline: float | int = 1
    spacing: float | int = 0.75
    shadow: float | int = 0
    alignment: int = 5
    bold: bool = True
    angle: float = 0.0
    borderstyle: int = 1
    italic: bool = False
    underline: bool = False


    def return_style(self):
        # convert Colors to pysub2.Color if in string format
        self.primarycolor = hex_to_pysub2_color(self.primarycolor) if type(self.primarycolor) is str else self.primarycolor
        self.backcolor = hex_to_pysub2_color(self.backcolor) if type(self.backcolor) is str else self.backcolor
        self.secondarycolor = hex_to_pysub2_color(self.secondarycolor) if type(self.secondarycolor) is str else self.secondarycolor
        self.outlinecolor= hex_to_pysub2_color(self.outlinecolor) if type(self.outlinecolor) is str else self.outlinecolor
        self.tertiarycolor = hex_to_pysub2_color(self.tertiarycolor) if type(self.tertiarycolor) is str else self.tertiarycolor
        # return the pysub2 style
        return pysubs2.SSAStyle(
            fontname=self.fontname,
            fontsize=self.fontsize,
            primarycolor=self.primarycolor,
            backcolor=self.backcolor,
            secondarycolor=self.secondarycolor,  # Black for border/shadow
            outlinecolor=self.outlinecolor,  # Black outline
            tertiarycolor=self.tertiarycolor,
            outline=self.outline,
            spacing=self.spacing,
            shadow=self.shadow,
            alignment=pysubs2.Alignment(self.alignment),
            bold=self.bold,
            angle=self.angle,
            borderstyle=self.borderstyle,
            italic=self.italic,
            underline=self.underline
        )

def is_drawing_line(line):
    """
    Determines if an ASS subtitle line contains a drawing or text.

    Args:
        line (str): A line from an ASS subtitle file

    Returns:
        bool: True if the line contains a drawing, False if it's text

    Examples:
        >>> is_drawing_line("Can I help you? Hmm?")
        False
        >>> is_drawing_line("{\\an7\\pos(640,678)\\p1}m -122 18.45 b -133.05...")
        True
    """
    # Check if line contains drawing mode tag \p1 or higher
    # \p0 means text mode, \p1 or higher means drawing mode
    drawing_pattern = r'\\p[1-9]'

    if re.search(drawing_pattern, line):
        return True

    return False

def segment_subs(events:list):
    result = []
    for event in events:
        if not is_drawing_line(event.text):
            result.append([event])
        else:
            result[-1].append(event)
    return result

def separate_tags(text:str):
    styles = text[text.find('{')+1:text.find('}')]
    return styles

def filter_out_fad_tag(tag:str):
    tags = tag.split(r"\ "[0])
    for tag in tags:
        if tag.__contains__('fad'):
            return tag
    return None

def fix_fad_issue(events:list):
    # check which tags dont contain an fad and then insert the needed fad tag
    segments = segment_subs(events)
    results = list()
    for segment in segments:
        text_tag = separate_tags(segment[0].text)
        background_tags = [separate_tags(x.text) for x in segment[1:]]

        # only continue if a fad statement is present
        if text_tag.__contains__('fad'):
            fad_tag = filter_out_fad_tag(text_tag)
            new_background_events = list()

            # only insert where the fad tag is not present
            for i, background_tag in enumerate(background_tags):
                if not fad_tag in background_tag:

                    # insert the new tags into the original text
                    new_background_events.append(background_tag + '\\'+ fad_tag)

                    # insert the new tags into the original text
                    segment[i+1].text = segment[i+1].text.replace(background_tag, background_tag + '\\'+ fad_tag)


            results.append(new_background_events)
    return segments



class subs_builder():

    def __init__(self):
        # tracks the dimensions of the builded subs, which is important to rebuild the backgrounds, if present
        self.dimensions_tracker = list()

    def __call__(self, subs, type=None, return_depth:bool=False): # text_only
        if type == 'text_only':
            return self.text_only_build(subs)
        if type == 'return_only_highlighted_texts':
            return self.return_only_highlighted_texts(subs)
        else:
            return self.normal_build(subs, return_depth)

    def normal_build(self, subs:list, return_depth:bool=False):
        new_subs = list()
        depth = list()
        build_subs = [x() for x in subs]
        for sub in build_subs:
            if type(sub) == list:
                new_subs.extend(sub)
                depth.append(len(sub))
            else:
                depth.append(1)
                new_subs.append(sub)

        # Depth needed to reverse input
        if not return_depth:
            return new_subs
        else:
            return new_subs, depth

    def text_only_build(self, subs:list):
        return [sub.return_saa_event() for sub in subs]

    def return_only_highlighted_texts(self, subs:list):
        return_subs = list()
        for sub in subs:
            new_subs = sub.return_only_highlighted_texts()
            # tracks length of subs
            self.dimensions_tracker.append(len(new_subs))
            return_subs.extend(new_subs)
        return return_subs

@dataclasses.dataclass(repr=False, eq=False, order=False)
class advanced_SAA_Events(pysubs2.SSAEvent):

    #text_list:list = ()
    highlighted_texts: list = ()
    highlight_style: list = ()
    appear_style:list = ()
    fade_in:float = 0
    fade_out:float = 0
    backgrounds:list = ()

    @property
    def text_list(self):
        return self.text.split(' ')

    def add_highlight_entry(self, index_start:int, index_end:int, start:int, end:int):
        # replace the tuple
        self.highlighted_texts = [] if len(self.highlighted_texts) == 0 else self.highlighted_texts
        self.highlighted_texts.append([index_start, index_end, start, end])

    def generate_fade(self, len_subs:int):
        if len_subs == 1:
            return fr'{{\fad({self.fade_in},{self.fade_out})}}', ''
        else:
            return fr'{{\fad({self.fade_in},0)}}', fr'{{\fad(0,{self.fade_out})}}'

    def return_saa_event(self):
        return pysubs2.SSAEvent(text=self.text.strip(), start=self.start, end=self.end, style="MainStyle")


    def return_only_highlighted_texts(self):
        return_list = []
        for index_start, index_end, start, end in self.highlighted_texts:
            return_list.append(pysubs2.SSAEvent(text=' '.join(self.text_list[index_start:index_end+1]), start=start, end=end, style="MainStyle", layer=self.layer))
        return return_list

    def add_background(self, backgrounds:list):
        if self.backgrounds == ():
            self.backgrounds = [backgrounds]
        else:
            self.backgrounds.append(backgrounds)

    def __call__(self):
        return_subs = []
        if self.highlighted_texts != () and self.highlighted_texts != []:
            # If appear is true, replace the highlight styles
            if self.appear_style != () and self.appear_style != []:
                self.highlight_style = ['', self.appear_style[0]]

            # Build the subs
            for i, (index_start, index_end, start, end)  in enumerate(self.highlighted_texts):
                # build the text with hightlighting marks
                text = f'{' '.join(self.text_list[0:index_start])} {self.highlight_style[0]}{' '.join(self.text_list[index_start:index_end+1])}{self.highlight_style[1]} {' '.join(self.text_list[index_end+1:])}'
                return_subs.append(pysubs2.SSAEvent(text=text.strip(), start=start, end=end, style="MainStyle", layer=self.layer))

                # add backgrounds to the subs
                if self.backgrounds != ():
                    return_subs.extend(self.backgrounds[i])

        else:
            return_subs = [pysubs2.SSAEvent(text=self.text, start=self.start, end=self.end, style="MainStyle", layer=self.layer)]

        # apply fade and return
        if len(return_subs) == 1:
            return_subs[0].text = self.generate_fade(len(return_subs))[0] + return_subs[0].text
            return return_subs[0]
        else:
            fade = self.generate_fade(len(return_subs))
            return_subs[0].text = fade[0] + return_subs[0].text
            return_subs[-1].text = fade[1] + return_subs[-1].text
            return return_subs


import pysubs2
import fleep
from . import utils
import pathlib
import os

class Input_Output_Handler:

    def __init__(self, args_sub_edit):
        self.whisper = self.handle_whisper_import()

        self.model = args_sub_edit.whisper_model
        self.device = args_sub_edit.whisper_device
        self.refine = args_sub_edit.whisper_refine

        self.data_input = args_sub_edit.input
        self.data_output = args_sub_edit.output
        self.video = args_sub_edit.input_video

        self.duration, self.resolution = self.handle_duration_resolution()

    def handle_input(self):
        sub_file = None

        # 1. Handle Whisper Objects
        if isinstance(self.data_input, self.whisper.result.WhisperResult):
            subs_str = self.data_input.to_srt_vtt(None, segment_level=False, word_level=True)
            sub_file = pysubs2.SSAFile.from_string(subs_str)

        # 2. Handle Dictionaries/Lists (Whisper JSON)
        if isinstance(self.data_input, (dict, list)):
            sub_file = pysubs2.load_from_whisper(self.data_input)

        # 3. Handle Strings (Paths or Raw Text)
        if isinstance(self.data_input, str):
            if not os.path.isfile(self.data_input):
                raise FileNotFoundError(f'{self.data_input} is not a file')

            if self.data_input.endswith(('.srt', '.ass')):
                sub_file = pysubs2.load(self.data_input)

            if self.is_media_file(self.data_input):
                subs_str = self.whisper_transcribe(self.data_input)
                sub_file = pysubs2.SSAFile.from_string(subs_str)

        # Handle video and resolution logic
        # for some parts of the effects the PlayResX and Y has to be set in the ass file
        if self.video is not None:
            sub_file = self.handle_subfile_resolution(sub_file)

        if sub_file is not None:
            return sub_file
        else:
            raise TypeError('Invalid input type')

    def handle_output(self, sub_file:pysubs2.SSAFile):

        # 1. Handle Strings
        if isinstance(self.data_output, str):
            if self.data_output.endswith('ass'):
                sub_file.save(self.data_output)
                return None

            if self.video is not None and self.is_output_video_file(self.video):
                utils.add_subtitles_with_ffmpeg(self.video, self.data_output, sub_file)
                return None

            raise TypeError('Output format has to be a either ".ass" or a video type')

        # 2. Return subfile
        if self.data_output is None:
            return sub_file

        raise TypeError('Invalid Output type')

    def handle_whisper_import(self):
        try:
            import stable_whisper
            return stable_whisper
        except ImportError as e:
            print('Import error:', e)
            return None

    def is_media_file(self, file_path):
        try:
            with open(file_path, "rb") as f:
                info = fleep.get(f.read(128))
                return any(t in info.type for t in ['audio', 'video'])
        except (FileNotFoundError, IsADirectoryError):
            return False

    def is_video_file(self, file_path):
        try:
            with open(file_path, "rb") as f:
                info = fleep.get(f.read(128))
                return any(t in info.type for t in ['video'])
        except (FileNotFoundError, IsADirectoryError):
            return False

    def is_output_video_file(self, file_path):
        # Define a set of common video extensions
        video_extensions = ('.mp4', '.mkv', '.avi', '.mov', '.wmv', '.flv', '.webm', '.m4v', '.m4p', '.ogv')
        extension = pathlib.Path(file_path).suffix.lower()

        return extension in video_extensions

    def whisper_transcribe(self, path):
        model = self.whisper.load_model(self.model, device=self.device)
        result = model.transcribe(audio=path, verbose=None)
        if self.refine:
            model.refine(path, result, word_level=False, only_voice_freq=True, precision=0.05)
        r = result.to_srt_vtt(None, segment_level=False, word_level=True)
        return r

    def handle_duration_resolution(self):
        # 1. Audio or Video from input
        if self.is_media_file(self.data_input):
            return utils.get_duration_resolution(self.data_input)

        # 2. Resolution and duration from extra input video
        if self.video is not None:
            if self.is_video_file(self.video):
                return utils.get_duration_resolution(self.video)

        return None, None

    def handle_subfile_resolution(self, subfile:pysubs2.SSAFile):
        info = subfile.info

        # check if resolution is set
        if utils.is_subfile_resolution_set(subfile):
            # check if resolution matches
            if self.resolution == (info['PlayResX'], info['PlayResY']):
                return subfile

        # if it is not or wrongly set, just add them to the file
        subfile.info['PlayResX'] = self.resolution[0]
        subfile.info['PlayResY'] = self.resolution[1]
        return subfile

    def build_full_sub_file(self, string_subs: str, script_info: str):
        segments = string_subs.split('[')
        segments[1] = script_info[1:]
        segments = '['.join(segments)
        return pysubs2.SSAFile.from_string(segments)


from . import base
import tarfile
import tempfile
from io import BytesIO
import re
from pysubs2 import SSAFile

class ContainerWrapper(base.BaseWrapper):

    def __init__(self, container, verbose:bool=False):
        # Check for the install
        super().__init__()

        # Only continue if package is installed
        if self.installed:
            self.client = self.get_client()
            self.verbose = verbose
            self.container = container

    def __call__(self, command: list | str, workdir:str='/home'):
        # Check if container is running
        self.container_running()
        # Check if the command is in the right format
        if type(command) is not str:
            command = self.build_command(command)
        # add exec to command
        command = ['bash', '-c', command]
        exit_code, output = self.container.exec_run(command, workdir=workdir)
        if exit_code != 0:
            raise RuntimeError(f'Container run into the following error with exit code {exit_code}: \n{output.decode('utf-8')}')
        if self.verbose:
            if output != '' and output != '\n':
                text = output.decode("utf-8").strip()
                if text:
                    self.logger.info("\n%s\n%s\n%s", '', text,)
                    print("\n" + "-" * 20)
                    print(text)
                    print("-" * 20 + "\n")

    def container_running(self):
        # loop for waiting container to start?
        self.container.reload()

        # Check if it's running
        if self.container.status != 'running':
            raise RuntimeError('Container is not running. It needs to be started first.')

    def build_command(self, commands: list[str]):
        return ' && '.join(commands)

    def copy_needed_files(self, input_ass:str | SSAFile, fonts_path:list | str = None):

        # Add if fonts should be copied
        if fonts_path is None:
            if_copy_fonts = False
        else:
            if_copy_fonts = True
            if type(fonts_path) is str:
                fonts_path = [fonts_path]

        # Adding input ass and fonts to tar
        with tempfile.NamedTemporaryFile(suffix='.tar.gz', delete=True) as tmp:

            with tarfile.open(fileobj=tmp, mode='w:gz') as tar:
                # Only add fonts if they are wanted
                if if_copy_fonts:
                    for font in fonts_path:
                        if not font.endswith(".ttf"):
                            print(f'Font "{font}" is not a ttf file')
                        else:
                            tar.add(font, arcname=f'fonts/{font.split("/")[-1]}')
                # Add input ass

                # if input is an ssafile object
                if type(input_ass) == SSAFile:
                    with tempfile.NamedTemporaryFile(suffix='.ass', delete=True) as ass:
                        input_ass.save(ass.name)
                        tar.add(ass.name, arcname='input.ass')
                else:
                    tar.add(input_ass, arcname='input.ass')

            # Flush and seek back to the beginning
            tmp.flush()
            tmp.seek(0)

            # import tar in docker
            self.container.put_archive('/home', tmp)
        # Put fonts in the right directory if they are needed
        if if_copy_fonts:
            command = [
                'mv /home/fonts/* /root/.local/share/fonts/',
                'rm -rf /home/fonts'
            ]
            self(command)

    def install_packages(self, packages:list[str]):
        command = ['apt-get update', f'apt-get install -y --no-install-recommends {' '.join(packages)}', 'rm -rf /var/lib/apt/lists/*']
        self(command)

    def retrieve_file(self, path:str, return_content:bool=False):
        bits, stat= self.container.get_archive(path)
        if return_content:
            buffer = BytesIO()
            for chunk in bits:
                buffer.write(chunk)
            buffer.seek(0)

            # Decode safely (ignore undecodable bytes)
            content = buffer.read().decode('utf-8', errors='ignore')

            # Remove BOMs and nulls
            content = content.lstrip('\ufeff\x00').rstrip('\x00')

            # Remove only *non-printable* control characters, except \n, \r, \t
            content = re.sub(r'[^\x20-\x7E\n\r\t]+', '', content)

            # Final clean string and return
            return content.strip()

        else:
            return bits











import os
import pysubs2
from . import utils
from .utils import dprint, advanced_SAA_Events
import dataclasses

@dataclasses.dataclass(kw_only=True)
class highlight_args(utils.args_styles):
    """
        Configuration arguments for subtitle text highlighting.

        Inherits visual styling properties from `utils.args_styles` and adds
        specific constraints for the highlighting engine.

        Attributes:
            highlight_word_max (int | None): The maximum number of words to be
                highlighted simultaneously in a single subtitle event.
                Defaults to 0 (no limit or disabled).
        """
    highlight_word_max: int | None = 0

    def replace_main_style(self, main_style: pysubs2.SSAStyle):
        self.fontname = self.fontname if self.fontname is not None else main_style.fontname
        self.fontsize = self.fontsize if self.fontsize is not None else main_style.fontsize
        self.primarycolor = self.primarycolor if self.primarycolor is not None else main_style.primarycolor
        self.backcolor = self.backcolor if self.backcolor is not None else main_style.backcolor
        self.secondarycolor = self.secondarycolor  if self.secondarycolor is not None else main_style.secondarycolor
        self.outlinecolor = self.outlinecolor if self.outlinecolor is not None else main_style.outlinecolor
        self.tertiarycolor = self.tertiarycolor if self.tertiarycolor is not None else main_style.tertiarycolor
        self.outline = self.outline if self.outline is not None else main_style.outline
        self.spacing = self.spacing if self.spacing is not None else main_style.spacing
        self.shadow = self.shadow if self.shadow is not None else main_style.shadow
        self.alignment = self.alignment if self.alignment is not None else main_style.alignment
        self.bold = self.bold if self.bold is not None else main_style.bold
        self.angle = self.angle if self.angle is not None else main_style.angle
        self.borderstyle = self.borderstyle if self.borderstyle is not None else main_style.borderstyle
        self.italic = self.italic  if self.italic is not None else main_style.italic
        self.underline = self.underline if self.underline is not None else main_style.underline


class Highlighter:

    def __init__(self, args:highlight_args, main_style: pysubs2.SSAStyle, subtitle_type:str):
        self.args = args
        self.args.replace_main_style(main_style)

        self.highlight_word_min = args.highlight_word_max

        # color, background, are possible values

        self.highlight_style = [r'{\rHighlight}', r'{\r}']

        self.subtitle_type = subtitle_type

    def __call__(self, cur_word: str, start, end, all_subs:list, sub_list: list):
        return_subs = list()
        highlighted_words = ''
        progress_in_words = ''
        cur_word += ' '
        last_index = 0

        sub_event = advanced_SAA_Events(text=cur_word, start=start, end=end, style="MainStyle", highlight_style=self.highlight_style)

        #dprint(sub_list)

        for i, sub in enumerate(sub_list):
            last_iteration = len(sub_list) - 1 == i

            if (self.highlight_word_min < len(highlighted_words) + len(sub.text)) or last_iteration:
                highlighted_words += f' {sub.text}'
                highlighted_words = highlighted_words.strip()

                progress_in_words = progress_in_words.strip()

                if not last_iteration:
                    end_time = sub_list[i + 1].start
                else:
                    end_time = end

                if start is None:
                    start = sub.start

                #return_subs.append(pysubs2.SSAEvent(start=start, end=end_time, text=new_cur_word.strip(), style="MainStyle"))
                sub_event.add_highlight_entry(last_index, i, start, end_time)

                last_index = i + 1
                highlighted_words = ''
                start = None
                progress_in_words += f' {sub.text}'
            else:
                highlighted_words += f' {sub.text}'
                progress_in_words += f' {sub.text}'
                if start is None:
                    start = sub.start

        #all_subs.append(return_subs)
        all_subs.append(sub_event)
        return all_subs

    def return_highlighted_style(self, style:pysubs2.SSAStyle):
        return self.args.return_style()

    def background_back(self):
        # return [r'{\3c&H000000&\4c&H0000FF&\4a&H40&\bord5\shad0\be1}', r'{\r}']
        return [r'{\c&H000000&\3c&HFFFF00&\bord5}', r'{\r}']  # {\c&H000000&\3c&HFFFF00&\bord8}

    def _replace(self, to_replace:str, new_word:str, progress_in_word:str):
        to_replace = to_replace.replace(progress_in_word, '', 1)
        if new_word[0] == ' ':
            _new = fr' {self.highlight_style[0]}{new_word.strip()}{self.highlight_style[1]} '
        else:
            _new = fr'{self.highlight_style[0]}{new_word.strip()}{self.highlight_style[1]} '

        highlighted_part = to_replace.replace(new_word, _new, 1)
        return progress_in_word + highlighted_part

    def highlight_background(self, subs):
        return_subs = list()
        background_style = self.background_back()
        for sub in subs:
            copy_subs = [x.copy() for x in sub]
            sub = self.replace_junk_styles(sub)
            for copy_sub in copy_subs:
                copy_sub.layer = 1
                for i, style in enumerate(background_style):
                    copy_sub.text = copy_sub.text.replace(self.highlight_style[i], style)
            return_subs.append(sub)
            return_subs.append(copy_subs)
        return return_subs

    def replace_junk_styles(self, subs:list):
        for sub in subs:
            for x in self.highlight_style:
                sub.text = sub.text.replace(x, '')
        return subs


from SubTextHighlight import docker_wrapper
import pysubs2
from pathlib import Path
import os

UPDATE_GOLDEN = os.environ.get("UPDATE_GOLDEN", "false").lower() == "true"

def test_docker():
    base_path = Path("")
    arial_path = base_path / "input" / "Arial-Rounded-MT-Bold-Bold.ttf"
    petemoss_path = base_path / "input" / "Petemoss-Regular.ttf"
    ass_path = base_path / "input" / "docker_sub_file.ass"
    expected_file = base_path / "expected" / "docker_sub_file.ass"

    dw = docker_wrapper.main.DockerWrapper()
    args_border = docker_wrapper.base.args_border(
        fonts_path=[str(arial_path), str(petemoss_path)],
    )

    input_ass = pysubs2.load(str(ass_path))
    output_file = dw(
        input_ass=input_ass,
        _traceback=True,
        cleanup=True,
        verbose=True,
        args_border=args_border,
    )
    if UPDATE_GOLDEN:
        output_file.save(str(expected_file))
        assert True
    else:
        expected_sub_file = pysubs2.load(str(expected_file))
        assert expected_sub_file.to_string('ass') == output_file.to_string('ass')





import os
import logging
import pysubs2
import subprocess

class BaseWrapper:

    def __init__(self,
                 force_install:bool = False):
        # Check for the install
        self.installed, self.docker = self.check_docker_installation()
        self.force_install = force_install
        self.logger = logging.getLogger("__main__")

        # Ask if package is not installed
        if force_install and not self.installed:
            self.install_docker_package()
        elif not self.installed:
            self.ask_install_docker_package()

    def check_docker_installation(self):
        try:
            import docker
            return True, docker
        except ImportError:
            return False, None

    def ask_install_docker_package(self):

        # Ask for confirmation to download the package
        print('For this Part of the script the Docker package must be installed')
        answer = ''
        while answer not in ['y', 'n']:
            answer = input('y/n:').lower()

        # Review answer
        if answer == 'y':
            self.install_docker_package()

        elif answer == 'n':
            pass

    def install_docker_package(self):
        try:
            subprocess.run(['pip', 'install', 'docker'], encoding='utf-8')
        except Exception as e:
            print(e)
            print('Could not install docker package')

        # verify docker install
        self.installed, self.docker = self.check_docker_installation()
        if not self.installed:
            raise ImportError('Docker could not be imported')


    def get_client(self):
        client = self.docker.from_env()
        return client


def pysub2_color_to_hex(color:pysubs2.Color):
    """
    Convert pysub2.Color object to hex format with separate alpha.

    Args:
        color: pysub2.Color object with r, g, b, a properties

    Returns:
        Tuple of (hex_color, alpha) where:
        - hex_color is a string in format #RRGGBB
        - alpha is the alpha value as an integer (0-255)
    """
    # Access the color properties from the Color object
    red = color.r
    green = color.g
    blue = color.b
    alpha = color.a

    # Convert to hex format (RGB only)
    hex_color = f"#{red:02X}{green:02X}{blue:02X}"

    return hex_color, alpha

class args_border:

    # container_run_func needs to take client and name as a input
    def __init__(self,
        offset:int = 6,
        radius:int = 6,
        transformy:int = 1,
        height_scaling:float = 1.2,
        color:pysubs2.Color | None = None,
        use_borders_as_highlight:bool = False,
        fonts_path: list | str | None = None,
        packages: list[str] | None = None,
        container_run_func : None=None,
        force_install:bool = False,
                 ):
        """
            Configuration for subtitle background borders or "box" styles.

            This class defines the geometry, color, and environment settings for
            rendering borders around subtitle text. It supports external rendering
            via containerized functions and custom package dependencies.

            Attributes:
                offset (int): The padding/offset of the border from the text.
                radius (int): The corner radius for rounded borders.
                transformy (int): Vertical transformation or shift for the border position.
                height_scaling (float): Multiplier to adjust the border height relative
                    to the text.
                bordercolor (str): Hex color code of the border (derived from input color).
                borderalpha (int): Alpha transparency level (derived from input color).
                use_borders_as_highlight (bool): If True, the border style is used to
                    indicate highlighted words.
                fonts_path (list | str | None): Path(s) to custom fonts required for
                    the border rendering.
                packages (list[str] | None): List of system or Python packages needed
                    inside the rendering container.
                container_run_func (Callable | None): A given function to return a
                    docker container with customized parameters. It gets the
                    docker client and name of the image as inputs.
                force_install (bool): If true, automatically install the docker package
                    as well as the necessary image. Otherwise, you are prompted to install.
        """
        self.offset:int = offset
        self.radius:int = radius
        self.transformy:int = transformy
        if color is not None:
            self.bordercolor, self.borderalpha = pysub2_color_to_hex(color)
        else:
            self.bordercolor = "#FFFFFF"
            self.borderalpha = 0
        self.packages:list[str] = packages
        self.fonts_path:list[str] = fonts_path
        self.container_run_func = container_run_func
        self.use_borders_as_highlight = use_borders_as_highlight
        self.height_scaling:float = height_scaling
        self.force_install = force_install

    def __call__(self):
        return {
            "offset": self.offset,
            "radius": self.radius,
            "transformY": self.transformy,
            "heightscaling": self.height_scaling,
            "borderColor": self.bordercolor,
            "borderAlpha": self.borderalpha,
        }



from . import Highlight
from . import docker_wrapper
from . import utils
import copy

import pysubs2

class effects_args:

    def __init__(self,
        fade:tuple[float, float] = (0.0, 0.0), # first is fadeIn and second is fadeOut
        appear:bool = False,
        args_border:docker_wrapper.base.args_border | None = None,
                 ):
        """
            Configuration for visual subtitle effects and animations.

            This class defines how subtitles transition onto the screen and whether
            additional decorative elements, like borders, are applied.

            Attributes:
                fade_in_duration (float): Time in seconds for the subtitle to transition
                    from transparent to opaque.
                fade_out_duration (float): Time in seconds for the subtitle to transition
                    from opaque to transparent.
                appear (bool): If True, words in a sequence stay on screen as new ones
                    appear (cumulative display). If False, words typically replace
                    one another.
                args_border (docker_wrapper.base.args_border, optional): Configuration
                    object for border-specific styling and effects.
        """
        self.fade_in_duration = fade[0]
        self.fade_out_duration = fade[1]
        self.appear = appear
        self.args_border = args_border


class Effects:

    def __init__(self, args:effects_args):
        self.args = args

    def logic_highlighter(self, highlighter:Highlight.Highlighter, sample_highlighter:Highlight.Highlighter):
        # sees whether the highlighter already exists as it is needed for some effects
        if self.args.appear:
            # sets highlighter if it does not already exist
            if highlighter is not None:
                new_highlighter = highlighter
            else:
                new_highlighter = sample_highlighter
            return new_highlighter
        else:
            return highlighter

    def __call__(self, subs:list, sub_file:pysubs2.SSAFile):
        if self.args.fade_out_duration != 0 and self.args.fade_in_duration != 0:
            subs = self.fade(subs)
        if self.args.appear:
            subs = self.appear(subs)

        if self.args.args_border is not None:
            # Implement rounded borders
            subs = self.rounded_borders(subs, sub_file)
        return subs

    def fade(self, subs):
        for i, sub in enumerate(subs):
            sub.fade_in =  self.args.fade_in_duration
            sub.fade_out = self.args.fade_out_duration
        return subs

    def appear(self, subs:list):
        styles = (r'{\alpha&HFF}', '')
        for sub in subs:
            sub.appear_style = styles
        return subs

    def rounded_borders(self, subs:list, sub_file:pysubs2.SSAFile):
        builder = utils.subs_builder()

        # check whether res is set, else raise error
        if not utils.is_subfile_resolution_set(sub_file):
            raise RuntimeError('The subtitle file does not contain a Resolution. For the right scaling of the subtitles a input with a video resolution has to be set.')

        # Check if appear is active and if so throw an expectation
        if self.args.args_border.use_borders_as_highlight and self.args.appear:
            raise RuntimeError('Cant use borders as highlighted subtitles and the appear at the same time.')

        # Check if Borders are used as highlight and build part of the background (if one is given)
        use_subs, depth = builder(subs, return_depth=True)


        # if borders as highlight, replace highlight with appear
        if self.args.args_border.use_borders_as_highlight:
            highlight_style = subs[0].highlight_style
            for sub in use_subs:
                sub.text = r'{\alpha&HFF}' + sub.text
                sub.text = sub.text.replace(highlight_style[1], highlight_style[1] + r'{\alpha&HFF}')

        # make copy of ssafile
        sub_file_copy = copy.deepcopy(sub_file)
        sub_file_copy.events = use_subs

        #print(sub_file_copy.to_string('ass'))

        # start the docker wrapper and execute the script
        # only execute on the part, that becomes the background
        dw = docker_wrapper.main.DockerWrapper(self.args.args_border.force_install)
        output : pysubs2.SSAFile = dw(
            input_ass=sub_file_copy,
            args_border=self.args.args_border,
            _traceback=True,
            cleanup=True,
        )

        events = output.events

        # fix the fad tag issue
        segmented_subs = utils.fix_fad_issue(events)

        # layer the subs
        for sub in subs:
            sub.layer = 1

        # merge subs and backgrounds
        #subs[index].add_background(segment[1:])
        segment_index = 0
        for i, cur in enumerate(depth):
            for x in range(cur):
                subs[i].add_background(segmented_subs[segment_index][1:])
                segment_index += 1

        return subs


import os
import datetime
import pysubs2
from Cython.Build.Dependencies import join_path
from .utils import dprint, advanced_SAA_Events
from .Highlight import Highlighter, highlight_args
from .Effects import Effects, effects_args
from . import utils
import fleep
import stable_whisper
import dataclasses
from . import handler

# TODO: Update Example Video
# TODO: Package and distribute package

off_time = datetime.timedelta(seconds=0.025)

@dataclasses.dataclass(kw_only=True)
class sub_args(utils.args_styles):
    """
    Configuration for the core subtitle generation and transcription process.

    This class defines the input/output paths, the transcription engine
    settings (Whisper), and the logic for how text is segmented into
    subtitle events.

    Attributes:
        input (str | dict | list | WhisperResult): The source to process. Can be
            a file path, a pre-transcribed dictionary/list, or a
            `stable_whisper.WhisperResult` object.
        output (str | None): File path where the generated subtitle file will
            be saved.
        input_video (str | None): Path to the source video file. Used for
            resolution detection and potential burning of subtitles.
        subtitle_type (str): The segmentation strategy. Supported options:
            - 'one_word_only': Displays exactly one word at a time.
            - 'join': Groups words into segments up to `word_max`.
            - 'separate_on_period': Splits segments at sentence boundaries.
        word_max (int): Maximum word count per subtitle event.
            Note: This is ignored if `subtitle_type` is 'one_word_only'.
        add_time (float): Time offset (in seconds) to extend the duration
            of each subtitle segment.
        fill_sub_times (bool): If True, ensures there are no gaps between
            consecutive subtitle segments.
        whisper_model (str): The specific OpenAI Whisper model size or
            language variant (e.g., 'medium.en', 'large-v3').
        whisper_device (str): The hardware device for inference (e.g., 'cpu',
            'cuda', or 'mps').
        whisper_refine (bool): If True, uses `stable-whisper` refinement to
            improve timestamp precision using audio frequencies.
    """

    input: str | dict[str, any] | list[dict[str, any]] | stable_whisper.result.WhisperResult
    output: str | None
    input_video: str | None = None
    subtitle_type: str = 'one_word_only'  # one_word_only, join, separate_on_period, appear
    word_max: int = 11
    add_time: float = 0
    fill_sub_times: bool = True
    whisper_model: str = 'medium.en'
    whisper_device: str = 'cpu'
    whisper_refine: bool = False



class Subtitle_Edit:
    """
        The central engine for subtitle generation and stylistic processing.

        This class handles the end-to-end workflow of subtitle creation, including
        input interpretation, style application, formatting logic (e.g., word-level
        splitting), visual effects, and final file building.

        Attributes:
            args (sub_args): Configuration object for subtitle editing and paths.
            main_style (ass.Style): The base visual style for the subtitles.
            word_max (int): Maximum number of words allowed per subtitle event.
            subtitle_type (str): The formatting strategy ('one_word_only',
                'separate_on_period', or 'join').
            highlighter (Highlighter, optional): Instance responsible for text
                highlighting logic.
            effects (Effects, optional): Instance responsible for visual animations
                and advanced styling.
            builder (utils.subs_builder): Utility to compile final subtitle events.
        """

    def __init__(self,
                 args_sub_edit:sub_args,
                 args_highlight:highlight_args | None = None,
                 args_effects: effects_args | None = None,
                ):
        """
                Initializes the Subtitle_Edit class with configuration and styles.

                Args:
                    args_sub_edit (sub_args): Core configurations including input/output
                        paths and model settings.
                    args_highlight (highlight_args, optional): Settings for text
                        highlighting. Defaults to None.
                    args_effects (effects_args, optional): Settings for visual effects
                        and animations. Defaults to None.
        """

        # args
        self.args = args_sub_edit

        # Style
        self.main_style = self.args.return_style()

        # Needed Variables for the formatting
        self.word_max = self.args.word_max
        self.subtitle_type = self.args.subtitle_type
        self.add_time = self.args.add_time
        self.fill_sub_times = self.args.fill_sub_times

        # Set handler and builder
        self.builder = utils.subs_builder()
        self.Handler = handler.Input_Output_Handler(args_sub_edit)

        # Highlighters
        #self.args_highlight = args_highlight

        if args_highlight is None:
            self.highlighter = None
        else:
            self.highlighter = Highlighter(args_highlight, self.main_style, self.subtitle_type)

        # Effects
        if args_effects is None:
            self.effects = None
        else:
            sample_highlighter = Highlighter(highlight_args(), self.main_style, self.subtitle_type)
            self.effects = Effects(args_effects)
            self.highlighter = self.effects.logic_highlighter(self.highlighter, sample_highlighter)


    def __call__(self):
        sub_file = self.Handler.handle_input()
        sub_file.styles["MainStyle"] = self.main_style

        if self.highlighter is not None:
            sub_file.styles["Highlight"] = self.highlighter.return_highlighted_style(self.main_style)

        subs = sub_file.events

        # create subtitles
        if self.subtitle_type == 'one_word_only':
            subs = self.one_word_only(subs)
        elif self.subtitle_type == 'separate_on_period':
            subs = self.short_subtitles(subs)
        elif self.subtitle_type == 'join':
            subs = self.short_subtitles_no_separation(subs)
        else:
            raise ValueError('Unsupported subtitle_type, please use a supported option.')

        # shift time
        if self.add_time != 0:
            subs = self.shift_subs_time(subs)

        # edit

        if self.effects is not None:
            subs  = self.effects(subs, sub_file)

        # build and save
        subs = self.builder(subs)
        sub_file.events = subs
        return self.Handler.handle_output(sub_file)

    def add_subtitle(self, cur_word:str, index:int, start, end, all_subs:list, highlight_words:bool=False, sub_list:list=()):
        if highlight_words is True:
            return self.highlighter(cur_word, start, end, all_subs, sub_list)
        else:
            all_subs.append(advanced_SAA_Events(start=start, end=end, text=cur_word.strip(), style="MainStyle"))
            return all_subs

    def short_subtitles(self, subs:list):
        word_highlight = self.return_if_highlight()
        new_subs = list()
        cur_word = ''
        index = 1
        start_time, end_time = self.start_end_time(subs)
        cur_sub_list = []

        for i, sub in enumerate(subs):
            #dprint(new_subs)
            last_iteration = len(subs) - 1 == i

            if sub.text.__contains__('.') or sub.text.__contains__('?') or sub.text.__contains__('!') or sub.text.__contains__(',') or last_iteration:
                cur_word = cur_word + sub.text
                cur_sub_list.append(sub)

                cur_end = self.return_end_time_logic(last_iteration, end_time, subs, sub, i)

                new_subs = self.add_subtitle(cur_word, index, start_time, cur_end, new_subs, highlight_words=word_highlight, sub_list=cur_sub_list)

                if not last_iteration:
                    start_time = subs[i+1].start

                cur_word = ''
                cur_sub_list = []
            else:
                cur_word = cur_word + sub.text + ' '
                cur_sub_list.append(sub)

        return new_subs

    def one_word_only(self, subs:list):
        word_highlight = self.return_if_highlight()
        new_subs = list()
        index = 1
        start_time, end_time = self.start_end_time(subs)

        for i, sub in enumerate(subs):
            last_iteration = len(subs) - 1 == i

            if not last_iteration:
                if self.fill_sub_times:
                    cur_end = subs[i+1].start
                else:
                    cur_end = sub.end
            else:
                cur_end = end_time

            new_subs = self.add_subtitle(sub.text, index, start_time, cur_end, new_subs, highlight_words=word_highlight,)

            if not last_iteration:
                start_time = subs[i + 1].start

        return new_subs

    def short_subtitles_no_separation(self, subs:list):
        word_highlight = self.return_if_highlight()
        new_subs = list()
        cur_word = ''
        cur_sub_list = []
        index = 1
        start_time, end_time = self.start_end_time(subs)

        for i, sub in enumerate(subs):
            last_iteration = len(subs) - 1 == i
            cur_word = f'{cur_word} {sub.text}'.strip()
            cur_sub_list.append(sub)

            next_word_len = len(cur_word) if last_iteration else len(cur_word) + 1 + len(subs[i+1].text)
            if self.word_max < next_word_len or last_iteration:

                cur_end = self.return_end_time_logic(last_iteration, end_time, subs, sub, i)

                new_subs = self.add_subtitle(cur_word, index, start_time, cur_end, new_subs, highlight_words=word_highlight, sub_list=cur_sub_list)
                cur_word = ''
                cur_sub_list = []
                if not last_iteration:
                    start_time = subs[i + 1].start

        return new_subs


    def return_if_highlight(self):
        if self.highlighter is None:
            return False
        else:
            return True

    def start_end_time(self, subs:list):
        if not self.fill_sub_times:
            return subs[0].start, subs[-1].end
        else:
            if self.Handler.duration is not None:
                return pysubs2.make_time(s=0), pysubs2.make_time(s=self.Handler.duration)
            else:
                raise ValueError('For the argument "fill_sub_times" an video has to be inputted via input_video or the subtitles have to generated from a audio/video.')

    def return_end_time_logic(self, last_iteration:bool, end_time:int, subs:list, sub:pysubs2.SSAEvent, i:int):
        if last_iteration:
            return end_time
        else:
            if self.fill_sub_times:
                return subs[i + 1].start
            else:
                return sub.end


    def shift_subs_time(self, subs:list):
        add_time = self.add_time
        for i, sub in enumerate(subs):
            if type(sub) == list:
                for _sub in sub:
                    _sub.shift(s=add_time)
            else:
                sub.shift(s=add_time)
        return subs


def generate_subs_simple(
        input,
        output,
    ):
    """
        Generates one-word-only subtitles from a media file with refined timestamps.

        This is a convenience wrapper that validates the input file type,
        transcribes it using the 'one_word_only' strategy, applies Whisper
        refinement for high precision, and saves the final result to disk.

        Args:
            input (str): Path to the source audio or video file.
            output (str): Destination path where the subtitle file (e.g., .ass, .srt)
                will be saved.

        Raises:
            ValueError: If the input file is not identified as a valid audio or
                video format by the `fleep` library.
        """
    # check for right type
    with open(input, "rb") as file:
        info = fleep.get(file.read(128))
    if not info.type == ['audio'] or info.type == ['video']:
        raise ValueError('The subtitles have to be generated from a audio/video.')
    else:
        sub_arg = sub_args(
            input=input,
            output=None,
            subtitle_type='one_word_only',
            fill_sub_times=False,
            whisper_refine=True
        )
        sub_file: pysubs2.SSAFile = Subtitle_Edit(sub_arg)()
        sub_file.save(output)


